/*
 * Tlog-rec command-line parsing.
 *
 * vim:nomodifiable
 * 
 * ************************* WARNING! DO NOT EDIT! *************************
 * This file is automatically generated from rec_conf_cmd.c.m4.
 * *************************************************************************
 *
 * Copyright (C) 2016 Red Hat
 *
 * This file is part of tlog.
 *
 * Tlog is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Tlog is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with tlog; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <config.h>
#include <tlog/rec_conf_validate.h>
#include <tlog/rec_conf_cmd.h>
#include <tlog/json_misc.h>
#include <tlog/rc.h>
#include <tlog/misc.h>
#include <libgen.h>
#include <stdio.h>
#include <unistd.h>
#include <getopt.h>
#include <string.h>
#include <inttypes.h>
#include <assert.h>

static const char *tlog_rec_conf_cmd_help_fmt =
    "Usage: %1$s [OPTION...] [CMD_FILE [CMD_ARG...]]\n"
    "   or: %1$s -c [OPTION...] CMD_STRING [CMD_NAME [CMD_ARG...]]\n"
    "Start a shell and log terminal I/O.\n"
    "\n"
    "General options:\n"
    "    -h, --help                  Output a command-line usage message and exit\n"
    "    -v, --version               Output version information and exit\n"
    "    -s, --shell=SHELL           Spawn the specified SHELL\n"
    "    -l, --login                 Make the shell a login shell\n"
    "    -c, --command               Execute shell commands\n"
    "    --notice=TEXT               Print TEXT message before starting recording\n"
    "    --latency=SECONDS           Cache captured data SECONDS seconds before logging\n"
    "    --payload=BYTES             Limit encoded data to BYTES bytes\n"
    "    -w, --writer=STRING         Use STRING log writer (syslog/file, default syslog)\n"
    "\n"
    "Logged data set options:\n"
    "    --log-input[=BOOL]          Enable/disable logging user input\n"
    "    --log-output[=BOOL]         Enable/disable logging program output\n"
    "    --log-window[=BOOL]         Enable/disable logging terminal window size changes\n"
    "\n"
    "File writer options:\n"
    "    --file-path=FILE            Log to FILE file\n"
    "\n"
    "Syslog writer options:\n"
    "    --syslog-facility=STRING    Log with STRING syslog facility\n"
    "    --syslog-priority=STRING    Log with STRING syslog priority\n"
    "";

static tlog_grc
tlog_rec_conf_cmd_load_args(struct tlog_errs **perrs,
                            struct json_object *conf,
                            const char *help,
                            int argc, char **argv)
{
    /* Option codes */
    enum opt {
        OPT_HELP = 'h',
        OPT_VERSION = 'v',
        OPT_SHELL = 's',
        OPT_LOGIN = 'l',
        OPT_COMMAND = 'c',
        OPT_WRITER = 'w',
        OPT_NOTICE = 0x100,
        OPT_LATENCY,
        OPT_PAYLOAD,
        OPT_LOG_INPUT,
        OPT_LOG_OUTPUT,
        OPT_LOG_WINDOW,
        OPT_FILE_PATH,
        OPT_SYSLOG_FACILITY,
        OPT_SYSLOG_PRIORITY,
    };

    /* Description of short options */
    static const char *shortopts = ":h::v::s:l::c::w:";

    /* Description of long options */
    static const struct option longopts[] = {
        {
            .name = "help",
            .val = OPT_HELP,
            .has_arg = optional_argument,
        },
        {
            .name = "version",
            .val = OPT_VERSION,
            .has_arg = optional_argument,
        },
        {
            .name = "shell",
            .val = OPT_SHELL,
            .has_arg = required_argument,
        },
        {
            .name = "login",
            .val = OPT_LOGIN,
            .has_arg = optional_argument,
        },
        {
            .name = "command",
            .val = OPT_COMMAND,
            .has_arg = optional_argument,
        },
        {
            .name = "notice",
            .val = OPT_NOTICE,
            .has_arg = required_argument,
        },
        {
            .name = "latency",
            .val = OPT_LATENCY,
            .has_arg = required_argument,
        },
        {
            .name = "payload",
            .val = OPT_PAYLOAD,
            .has_arg = required_argument,
        },
        {
            .name = "log-input",
            .val = OPT_LOG_INPUT,
            .has_arg = optional_argument,
        },
        {
            .name = "log-output",
            .val = OPT_LOG_OUTPUT,
            .has_arg = optional_argument,
        },
        {
            .name = "log-window",
            .val = OPT_LOG_WINDOW,
            .has_arg = optional_argument,
        },
        {
            .name = "writer",
            .val = OPT_WRITER,
            .has_arg = required_argument,
        },
        {
            .name = "file-path",
            .val = OPT_FILE_PATH,
            .has_arg = required_argument,
        },
        {
            .name = "syslog-facility",
            .val = OPT_SYSLOG_FACILITY,
            .has_arg = required_argument,
        },
        {
            .name = "syslog-priority",
            .val = OPT_SYSLOG_PRIORITY,
            .has_arg = required_argument,
        },
        {
            .name = NULL
        }
    };

    tlog_grc grc;
    int opterr_orig;
    int optind_orig;
    int optcode;
    const char *optname;
    const char *optpath;
    int64_t val_int;
    struct json_object *val = NULL;
    struct json_object *args = NULL;
    int end;
    int i;

    /* May be unused if there are no INT parameters */
    (void)val_int;
    (void)end;

    /* Ask getopt_long to not print an error message */
    opterr_orig = opterr;
    opterr = 0;

    /* Ask getopt_long to start from the first argument */
    optind_orig = optind;
    optind = 1;

    /* While there are options */
    while ((optcode = getopt_long(argc, argv, 
                                  shortopts, longopts, NULL)) >= 0) {
        switch (optcode) {
        case OPT_HELP:
            optname = "-h/--help";
            optpath = "help";
            if (optarg == NULL ||
                strcasecmp(optarg, "yes") == 0 ||
                strcasecmp(optarg, "on") == 0 ||
                strcasecmp(optarg, "true") == 0) {
                val = json_object_new_boolean(true);
            } else if (strcasecmp(optarg, "no") == 0 ||
                       strcasecmp(optarg, "off") == 0 ||
                       strcasecmp(optarg, "false") == 0) {
                val = json_object_new_boolean(false);
            } else {
                tlog_errs_pushf(perrs,
                                "Invalid %s option value: %s,\n"
                                "expecting yes/on/true or no/off/false\n%s",
                                optname, optarg, help);
                grc = TLOG_RC_FAILURE;
                goto cleanup;
            }
            break;

        case OPT_VERSION:
            optname = "-v/--version";
            optpath = "version";
            if (optarg == NULL ||
                strcasecmp(optarg, "yes") == 0 ||
                strcasecmp(optarg, "on") == 0 ||
                strcasecmp(optarg, "true") == 0) {
                val = json_object_new_boolean(true);
            } else if (strcasecmp(optarg, "no") == 0 ||
                       strcasecmp(optarg, "off") == 0 ||
                       strcasecmp(optarg, "false") == 0) {
                val = json_object_new_boolean(false);
            } else {
                tlog_errs_pushf(perrs,
                                "Invalid %s option value: %s,\n"
                                "expecting yes/on/true or no/off/false\n%s",
                                optname, optarg, help);
                grc = TLOG_RC_FAILURE;
                goto cleanup;
            }
            break;

        case OPT_SHELL:
            optname = "-s/--shell";
            optpath = "shell";
            assert(optarg != NULL);
            if (optarg == NULL) {
                tlog_errs_pushf(perrs, "Option %s has no value", optname);
                grc = TLOG_RC_FAILURE;
                goto cleanup;
            }
            val = json_object_new_string(optarg);
            break;

        case OPT_LOGIN:
            optname = "-l/--login";
            optpath = "login";
            if (optarg == NULL ||
                strcasecmp(optarg, "yes") == 0 ||
                strcasecmp(optarg, "on") == 0 ||
                strcasecmp(optarg, "true") == 0) {
                val = json_object_new_boolean(true);
            } else if (strcasecmp(optarg, "no") == 0 ||
                       strcasecmp(optarg, "off") == 0 ||
                       strcasecmp(optarg, "false") == 0) {
                val = json_object_new_boolean(false);
            } else {
                tlog_errs_pushf(perrs,
                                "Invalid %s option value: %s,\n"
                                "expecting yes/on/true or no/off/false\n%s",
                                optname, optarg, help);
                grc = TLOG_RC_FAILURE;
                goto cleanup;
            }
            break;

        case OPT_COMMAND:
            optname = "-c/--command";
            optpath = "command";
            if (optarg == NULL ||
                strcasecmp(optarg, "yes") == 0 ||
                strcasecmp(optarg, "on") == 0 ||
                strcasecmp(optarg, "true") == 0) {
                val = json_object_new_boolean(true);
            } else if (strcasecmp(optarg, "no") == 0 ||
                       strcasecmp(optarg, "off") == 0 ||
                       strcasecmp(optarg, "false") == 0) {
                val = json_object_new_boolean(false);
            } else {
                tlog_errs_pushf(perrs,
                                "Invalid %s option value: %s,\n"
                                "expecting yes/on/true or no/off/false\n%s",
                                optname, optarg, help);
                grc = TLOG_RC_FAILURE;
                goto cleanup;
            }
            break;

        case OPT_NOTICE:
            optname = "--notice";
            optpath = "notice";
            assert(optarg != NULL);
            if (optarg == NULL) {
                tlog_errs_pushf(perrs, "Option %s has no value", optname);
                grc = TLOG_RC_FAILURE;
                goto cleanup;
            }
            val = json_object_new_string(optarg);
            break;

        case OPT_LATENCY:
            optname = "--latency";
            optpath = "latency";
            assert(optarg != NULL);
            if (optarg == NULL) {
                tlog_errs_pushf(perrs, "Option %s has no value", optname);
                grc = TLOG_RC_FAILURE;
                goto cleanup;
            }
            if (sscanf(optarg, "%" SCNd64 " %n", &val_int, &end) < 1 ||
                optarg[end] != 0 || val_int < 1) {
                tlog_errs_pushf(perrs, "Invalid %s option value: %s\n%s",
                                optname, optarg, help);
                grc = TLOG_RC_FAILURE;
                goto cleanup;
            }
            val = json_object_new_int64(val_int);
            break;

        case OPT_PAYLOAD:
            optname = "--payload";
            optpath = "payload";
            assert(optarg != NULL);
            if (optarg == NULL) {
                tlog_errs_pushf(perrs, "Option %s has no value", optname);
                grc = TLOG_RC_FAILURE;
                goto cleanup;
            }
            if (sscanf(optarg, "%" SCNd64 " %n", &val_int, &end) < 1 ||
                optarg[end] != 0 || val_int < 32) {
                tlog_errs_pushf(perrs, "Invalid %s option value: %s\n%s",
                                optname, optarg, help);
                grc = TLOG_RC_FAILURE;
                goto cleanup;
            }
            val = json_object_new_int64(val_int);
            break;

        case OPT_LOG_INPUT:
            optname = "--log-input";
            optpath = "log.input";
            if (optarg == NULL ||
                strcasecmp(optarg, "yes") == 0 ||
                strcasecmp(optarg, "on") == 0 ||
                strcasecmp(optarg, "true") == 0) {
                val = json_object_new_boolean(true);
            } else if (strcasecmp(optarg, "no") == 0 ||
                       strcasecmp(optarg, "off") == 0 ||
                       strcasecmp(optarg, "false") == 0) {
                val = json_object_new_boolean(false);
            } else {
                tlog_errs_pushf(perrs,
                                "Invalid %s option value: %s,\n"
                                "expecting yes/on/true or no/off/false\n%s",
                                optname, optarg, help);
                grc = TLOG_RC_FAILURE;
                goto cleanup;
            }
            break;

        case OPT_LOG_OUTPUT:
            optname = "--log-output";
            optpath = "log.output";
            if (optarg == NULL ||
                strcasecmp(optarg, "yes") == 0 ||
                strcasecmp(optarg, "on") == 0 ||
                strcasecmp(optarg, "true") == 0) {
                val = json_object_new_boolean(true);
            } else if (strcasecmp(optarg, "no") == 0 ||
                       strcasecmp(optarg, "off") == 0 ||
                       strcasecmp(optarg, "false") == 0) {
                val = json_object_new_boolean(false);
            } else {
                tlog_errs_pushf(perrs,
                                "Invalid %s option value: %s,\n"
                                "expecting yes/on/true or no/off/false\n%s",
                                optname, optarg, help);
                grc = TLOG_RC_FAILURE;
                goto cleanup;
            }
            break;

        case OPT_LOG_WINDOW:
            optname = "--log-window";
            optpath = "log.window";
            if (optarg == NULL ||
                strcasecmp(optarg, "yes") == 0 ||
                strcasecmp(optarg, "on") == 0 ||
                strcasecmp(optarg, "true") == 0) {
                val = json_object_new_boolean(true);
            } else if (strcasecmp(optarg, "no") == 0 ||
                       strcasecmp(optarg, "off") == 0 ||
                       strcasecmp(optarg, "false") == 0) {
                val = json_object_new_boolean(false);
            } else {
                tlog_errs_pushf(perrs,
                                "Invalid %s option value: %s,\n"
                                "expecting yes/on/true or no/off/false\n%s",
                                optname, optarg, help);
                grc = TLOG_RC_FAILURE;
                goto cleanup;
            }
            break;

        case OPT_WRITER:
            optname = "-w/--writer";
            optpath = "writer";
            {
                const char *list[] = {"syslog",
                                      "file"};
                size_t i;
                assert(optarg != NULL);
                if (optarg == NULL) {
                    tlog_errs_pushf(perrs, "Option %s has no value", optname);
                    grc = TLOG_RC_FAILURE;
                    goto cleanup;
                }
                for (i = 0;
                     i < TLOG_ARRAY_SIZE(list) && strcmp(optarg, list[i]) != 0;
                     i++);
                if (i >= TLOG_ARRAY_SIZE(list)) {
                    tlog_errs_pushf(perrs, "Invalid %s option value: %s\n%s",
                                    optname, optarg, help);
                    grc = TLOG_RC_FAILURE;
                    goto cleanup;
                }
            }
            val = json_object_new_string(optarg);
            break;

        case OPT_FILE_PATH:
            optname = "--file-path";
            optpath = "file.path";
            assert(optarg != NULL);
            if (optarg == NULL) {
                tlog_errs_pushf(perrs, "Option %s has no value", optname);
                grc = TLOG_RC_FAILURE;
                goto cleanup;
            }
            val = json_object_new_string(optarg);
            break;

        case OPT_SYSLOG_FACILITY:
            optname = "--syslog-facility";
            optpath = "syslog.facility";
            {
                const char *list[] = {"auth",
                                      "authpriv",
                                      "cron",
                                      "daemon",
                                      "ftp",
                                      "kern",
                                      "local0",
                                      "local1",
                                      "local2",
                                      "local3",
                                      "local4",
                                      "local5",
                                      "local6",
                                      "local7",
                                      "lpr",
                                      "mail",
                                      "news",
                                      "syslog",
                                      "user",
                                      "uucp"};
                size_t i;
                assert(optarg != NULL);
                if (optarg == NULL) {
                    tlog_errs_pushf(perrs, "Option %s has no value", optname);
                    grc = TLOG_RC_FAILURE;
                    goto cleanup;
                }
                for (i = 0;
                     i < TLOG_ARRAY_SIZE(list) && strcmp(optarg, list[i]) != 0;
                     i++);
                if (i >= TLOG_ARRAY_SIZE(list)) {
                    tlog_errs_pushf(perrs, "Invalid %s option value: %s\n%s",
                                    optname, optarg, help);
                    grc = TLOG_RC_FAILURE;
                    goto cleanup;
                }
            }
            val = json_object_new_string(optarg);
            break;

        case OPT_SYSLOG_PRIORITY:
            optname = "--syslog-priority";
            optpath = "syslog.priority";
            {
                const char *list[] = {"emerg",
                                      "alert",
                                      "crit",
                                      "err",
                                      "warning",
                                      "notice",
                                      "info",
                                      "debug"};
                size_t i;
                assert(optarg != NULL);
                if (optarg == NULL) {
                    tlog_errs_pushf(perrs, "Option %s has no value", optname);
                    grc = TLOG_RC_FAILURE;
                    goto cleanup;
                }
                for (i = 0;
                     i < TLOG_ARRAY_SIZE(list) && strcmp(optarg, list[i]) != 0;
                     i++);
                if (i >= TLOG_ARRAY_SIZE(list)) {
                    tlog_errs_pushf(perrs, "Invalid %s option value: %s\n%s",
                                    optname, optarg, help);
                    grc = TLOG_RC_FAILURE;
                    goto cleanup;
                }
            }
            val = json_object_new_string(optarg);
            break;

        case ':':
            for (i = 0; i < (int)TLOG_ARRAY_SIZE(longopts); i++) {
                if (longopts[i].val == optopt) {
                    break;
                }
            }
            if (i < (int)TLOG_ARRAY_SIZE(longopts)) {
                if (longopts[i].val < 0x100) {
                    tlog_errs_pushf(perrs,
                                    "Option -%c/--%s argument is missing\n%s",
                                    longopts[i].val, longopts[i].name, help);
                } else {
                    tlog_errs_pushf(perrs,
                                    "Option --%s argument is missing\n%s",
                                    longopts[i].name, help);
                }
            } else {
                tlog_errs_pushf(perrs,
                                "Option -%c argument is missing\n%s",
                                optopt, help);
            }
            grc = TLOG_RC_FAILURE;
            goto cleanup;

        case '?':
            if (optopt == 0) {
                tlog_errs_pushf(perrs,
                                "Unknown option encountered\n%s",
                                help);
            } else {
                tlog_errs_pushf(perrs,
                                "Unknown option encountered: -%c\n%s",
                                optopt, help);
            }
            grc = TLOG_RC_FAILURE;
            goto cleanup;

        default:
            tlog_errs_pushf(perrs, "Unknown option code: %d", optcode);
            grc = TLOG_RC_FAILURE;
            goto cleanup;
        }

        if (val == NULL) {
            grc = TLOG_GRC_ERRNO;
            tlog_errs_pushc(perrs, grc);
            tlog_errs_pushf(perrs, "Failed creating %s option value", optname);
            goto cleanup;
        }
        grc = tlog_json_object_object_add_path(conf, optpath, val);
        if (grc != TLOG_RC_OK) {
            tlog_errs_pushc(perrs, grc);
            tlog_errs_pushf(perrs, "Failed storing %s option value", optname);
            goto cleanup;
        }
        val = NULL;
    }

    /* Add other arguments */
    args = json_object_new_array();
    if (args == NULL) {
        grc = TLOG_GRC_ERRNO;
        tlog_errs_pushc(perrs, grc);
        tlog_errs_pushs(perrs, "Failed creating positional argument list");
        goto cleanup;
    }
    for (i = 0; optind < argc; i++, optind++) {
        val = json_object_new_string(argv[optind]);
        if (val == NULL) {
            grc = TLOG_GRC_ERRNO;
            tlog_errs_pushc(perrs, grc);
            tlog_errs_pushs(perrs, "Failed creating argument value");
            goto cleanup;
        }
        if (json_object_array_put_idx(args, i, val) < 0) {
            grc = TLOG_GRC_ERRNO;
            tlog_errs_pushc(perrs, grc);
            tlog_errs_pushf(perrs, "Failed storing argument #%d", i);
            goto cleanup;
        }
        val = NULL;
    }
    grc = tlog_json_object_object_add_path(conf, "args", args);
    if (grc != TLOG_RC_OK) {
        tlog_errs_pushc(perrs, grc);
        tlog_errs_pushs(perrs, "Failed storing argument list");
        goto cleanup;
    }
    args = NULL;

    /* Validate the result */
    grc = tlog_rec_conf_validate(perrs, conf, TLOG_CONF_ORIGIN_ARGS);
    if (grc != TLOG_RC_OK) {
        tlog_errs_pushs(perrs, "Failed to validate loaded configuration");
        goto cleanup;
    }

    grc = TLOG_RC_OK;

cleanup:
    json_object_put(args);
    json_object_put(val);
    optind = optind_orig;
    opterr = opterr_orig;
    return grc;
}

tlog_grc
tlog_rec_conf_cmd_load(struct tlog_errs **perrs,
                       char **phelp, struct json_object **pconf,
                       int argc, char **argv)
{
    tlog_grc grc;
    char *progpath = NULL;
    const char *p;
    char *progname = NULL;
    char *help = NULL;
    struct json_object *conf = NULL;
    struct json_object *val = NULL;

    assert(phelp != NULL);
    assert(pconf != NULL);
    assert(argv != NULL);

    /* Create empty configuration */
    conf = json_object_new_object();
    if (conf == NULL) {
        grc = TLOG_GRC_ERRNO;
        tlog_errs_pushc(perrs, grc);
        tlog_errs_pushs(perrs, "Failed creating configuration object");
        goto cleanup;
    }

    /* Extract program name, noting login dash prefix */
    progpath = strdup(argv[0]);
    if (progpath == NULL) {
        grc = TLOG_GRC_ERRNO;
        tlog_errs_pushc(perrs, grc);
        tlog_errs_pushs(perrs, "Failed allocating a copy of program path");
        goto cleanup;
    }
    p = basename(progpath);
    if (p[0] == '-') {
        p++;
        val = json_object_new_boolean(true);
        if (val == NULL) {
            grc = TLOG_GRC_ERRNO;
            tlog_errs_pushc(perrs, grc);
            tlog_errs_pushs(perrs, "Failed creating login flag");
            goto cleanup;
        }
        grc = tlog_json_object_object_add_path(conf, "login", val);
        if (grc != TLOG_RC_OK) {
            tlog_errs_pushc(perrs, grc);
            tlog_errs_pushs(perrs, "Failed storing login flag");
            goto cleanup;
        }
        val = NULL;
    }
    progname = strdup(p);
    if (progname == NULL) {
        grc = TLOG_GRC_ERRNO;
        tlog_errs_pushc(perrs, grc);
        tlog_errs_pushs(perrs, "Failed allocating program name");
        goto cleanup;
    }

    /* Extract options and positional arguments */
    if (asprintf(&help, tlog_rec_conf_cmd_help_fmt, progname) < 0) {
        grc = TLOG_GRC_ERRNO;
        tlog_errs_pushc(perrs, grc);
        tlog_errs_pushs(perrs, "Failed formatting help message");
        goto cleanup;
    }
    grc = tlog_rec_conf_cmd_load_args(perrs, conf, help, argc, argv);
    if (grc != TLOG_RC_OK) {
        tlog_errs_pushs(perrs,
                        "Failed extracting configuration "
                        "from options and arguments");
        goto cleanup;
    }

    /* Validate the result */
    grc = tlog_rec_conf_validate(perrs, conf, TLOG_CONF_ORIGIN_ARGS);
    if (grc != TLOG_RC_OK) {
        tlog_errs_pushs(perrs, "Validation of loaded configuration failed");
        goto cleanup;
    }

    *phelp = help;
    help = NULL;
    *pconf = conf;
    conf = NULL;
    grc = TLOG_RC_OK;

cleanup:
    free(help);
    free(progname);
    free(progpath);
    json_object_put(val);
    json_object_put(conf);
    return grc;
}
